# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Evaluation metrics."""
import dataclasses
import logging
from lsmmdma.mmdma_functions import compute_sqpairwise_distances
import numpy as np
from pykeops.torch import LazyTensor
import scipy
import torch


@dataclasses.dataclass()
class OutputSupervised():
  foscttm: float
  top1: float
  top5: float


# Supervised evaluation.
@dataclasses.dataclass()
class SupervisedEvaluation():
  """Computes evaluation metrics, knowing the ground truth alignment.

  Three metrics are being computed:
  - FOSCTTM (Fraction Of Samples Closer Than the True Match): it gives the
  average number of samples from one view that are closer to the true match
  in the other view, in the learned low dimensional space. As this metrics is
  not symmetrical if we exchange the first and second views, the results are
  averaged calculating the FOSCTTM in both directions.
  - topK: it computes the fraction of samples from one view correctly
  assigned among top K% nearest neighbours of the true match from the other
  view. As this metrics is not symmetrical if we exchange the first and
  second views, the results are averaged calculating top1 in both directions.
  K is set to be 1 and 5 by default.
  """
  ground_truth_alignment: np.ndarray
  device: torch.device

  def compute_all_evaluation(
      self,
      first_view: torch.FloatTensor,
      second_view: torch.FloatTensor
      ) -> OutputSupervised:
    """Computes evaluation measures, knowing the ground truth alignment."""
    second_view_aligned = second_view[self.ground_truth_alignment]

    n = first_view.shape[0]
    try:
      foscttm = self._foscttm(first_view, second_view_aligned)
    except:
      logging.warning(
          'FOSCTTM was not computed and most likely led to an OOM issue.')
      foscttm = -1
    if n > 100:
      top1 = self._topx_keops(
          first_view, second_view_aligned, topk=int(1 / 100 * n))
    else:
      logging.warning(
          'Top1 can not be computed with a number of samples <100.')
      top1 = -1
    if n > 20:
      top5 = self._topx_keops(
          first_view, second_view_aligned, topk=int(5 / 100 * n))
    else:
      logging.warning(
          'Top5 can not be computed with a number of samples <20.')
      top5 = -1
    return OutputSupervised(foscttm, top1, top5)

  def _foscttm(
      self,
      first_view: torch.FloatTensor,
      second_view: torch.FloatTensor
      ) -> float:
    """Computes the fraction of samples closer to the true match based on
    squared euclidean distances between samples."""
    # TODO(lpapaxanthos): Memory efficient FOSCTTM.
    n = first_view.shape[0]
    # Assumes the views are aligned.
    distances = compute_sqpairwise_distances(first_view, second_view)
    fraction = (
        torch.sum(distances < torch.diag(distances))
        + torch.sum(torch.t(distances) < torch.diag(distances))
        ) / (2 * n * (n - 1))
    return fraction.item()

  def _topx_keops(
      self,
      first_view: torch.FloatTensor,
      second_view: torch.FloatTensor,
      topk: int = 1
      ) -> float:
    """Computes fraction of samples correctly assigned among topk nearest
    neighbours."""
    def get_count_knn(distance: LazyTensor, dim: int = 0):
      # Grid <-> Samples, (M**2, K) integer tensor.
      indknn = distance.argKmin(topk, dim=dim)
      frac = indknn - torch.arange(n).reshape(-1, 1).to(self.device)
      return torch.count_nonzero(frac == 0).item()
    n = first_view.shape[0]
    first_view_i = LazyTensor(first_view[:, None, :])  # (M**2, 1, 2)
    second_view_j = LazyTensor(second_view[None, :, :])  # (1, N, 2)
    # (M**2, N) symbolic matrix of squared distances.
    distance_ij = ((first_view_i - second_view_j) ** 2).sum(-1)
    count0 = get_count_knn(distance_ij, dim=0)
    count1 = get_count_knn(distance_ij, dim=1)
    return (count0 + count1) / (2 * n)



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generates data with varying number of samples and features."""
import os
import random
from typing import Tuple, Union

import numpy as np
import torch
from tensorflow.io import gfile


def generate_data(
    n_sample: int,
    p_feature: int,
    random_seed: int = 4,
    random_seed_ext: int = 5,
    simulation: str = 'branch',
    ) -> Union[Tuple[np.ndarray], Tuple[torch.Tensor]]:
  """Generates two views following the shape of a branch in their latent space.

  Arguments:
    n_sample: number of samples.
    p_feature: number of features.
    random_seed: seed for random.
    random_seed_ext: seed for numpy.
    simulation: str, latent structure for the input data. The latent structure
      can either be a 'branch' (two elongated Gaussian point clouds
      perpendicular to each other) or a 'triangle' (three elongated Gaussian
      point clouds forming a triangle).

  Returns:
    tuple, first view, second view and corresponding indices from the first
      view.
  """
  latent_dim = 2
  # TODO(lpapaxanthos): allow for different number of features.
  p_features1 = p_feature
  p_features2 = p_feature
  noise = 0.05

  if simulation == 'triangle':
    means = [(0, 1), (3.5, 4.33), (-3.5, 4.33)]
    covs = [[[8, 0], [0, 0.05]],
            [[1.7, -1.5], [-1.5, 1.7]],
            [[1.7, 1.5], [1.5, 1.7]]]
    n1 = int(n_sample / 4)
    shapes = [2 * n1, n1, n1 - 3 * n1]
    rot = None
  elif simulation == 'branch':
    means = [[0, 0], [0, 8]]
    covs = [[[20, 0], [0, 1]], [[1, 0], [0, 10]]]
    n1 = int(n_sample / 4 * 3)
    shapes = [n1, n_sample - n1]
    rot = np.array([[np.cos(np.pi/4.), -np.sin(np.pi/4.)],
                    [np.sin(np.pi/4.), np.cos(np.pi/4.)]])
  else:
    raise ValueError('Value for simulation is not supported')

  length = len(means)

  random.seed(random_seed)
  np.random.seed(random_seed_ext)
  axis = [0] * length
  for i in range(length):
    axis[i] = np.random.multivariate_normal(
        mean=np.array(means[i]), cov=np.array(covs[i]), size=shapes[i])

  latent_array = np.concatenate((list(axis[i] for i in range(length))))
  if rot is not None:
    latent_array = np.dot(latent_array, rot)
  latent_array = (latent_array - latent_array.mean(0)) / latent_array.std(0)

  noise_first_view = np.random.normal(
      size=(latent_dim, p_features1))
  noise_second_view = np.random.normal(
      size=(latent_dim, p_features2))
  first_view = (torch.FloatTensor(
      np.dot(latent_array, noise_first_view) + np.random.normal(
          size=(n_sample, p_features1)) * noise))
  second_view = (torch.FloatTensor(
      np.dot(latent_array, noise_second_view) + np.random.normal(
          size=(n_sample, p_features2)) * noise))

  rd_vec = np.random.choice(n_sample, n_sample, replace=False)
  first_view = first_view[rd_vec]

  return first_view, second_view, rd_vec


def load(input_dir: str, filename: str) -> np.ndarray:
  with gfile.GFile(
      os.path.join(input_dir, filename), 'rb') as my_file:
    try:
      data = np.load(my_file)
    except:
      data = np.loadtxt(my_file)
  return data



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Main module that generates the data and trains the model.

Instructions:

1. To run the algorithm on simulated data from data_pipeline.py:

python3 -m lsmmdma.main --output_dir outdir \
--data branch --n 300 --p 400 \
--k 4 --ns 5 \
--e 1001 --d 5 --nr 100 --ne 100 --keops True --m dual --pca 100 \
--lr 1e-5 --l1 1e-4 --l2 1e-4 --s 1.0 --init 'uniform'

2. To run the algorithm on user input data, in the form n_sample x p_feature.
--data should be '' (default value) and --kernel should be False. The
argument --keops can be True or False, --mode can be 'dual' or 'primal'.

python3 -m lsmmdma.main --input_dir datadir --output_dir outdir \
--input_fv my_data_1 --input_sv my_data_2 --kernel False \
--k 4 --ns 5 \
--e 1001 --d 5 --nr 100 --ne 100 --keops True --m dual --pca 100 \
--lr 1e-5 --l1 1e-4 --l2 1e-4 --s 1.0 --init 'uniform'

3. To run the algorithm on user kernel data, in the form n_sample x n_sample.
--data should be '' (default value) and --kernel should be True. The
argument --keops can be True or False, --mode can only be `dual`.

python3 -m lsmmdma.main --inputdir datadir --output_dir outdir \
--input_fv my_data_1 --input_sv my_data_2 --kernel True \
--k 4 --ns 5 \
--e 1001 --d 5 --nr 100 --ne 100 --keops True --m dual --pca 100 \
--lr 1e-5 --l1 1e-4 --l2 1e-4 --s 1.0 --init 'uniform'
"""
import os

from absl import app
from absl import flags
from absl import logging
import dataclasses
from lsmmdma import train
from lsmmdma.data import checkpointer
from lsmmdma.data import data_pipeline
from lsmmdma.metrics import SupervisedEvaluation
import numpy as np
import torch
import tensorflow as tf
from tensorflow.io import gfile

# Flags for input and output.
flags.DEFINE_string('output_dir', '', 'Output directory.')
flags.DEFINE_string('input_dir', '', 'Output directory.')
flags.DEFINE_string(
    'input_fv', '', 'Input first view, can be point cloud or kernel.')
flags.DEFINE_string(
    'input_sv', '', 'Input second view, can be point cloud or kernel.')
flags.DEFINE_string('rd_vec', '', 'Permutation of first view.')
flags.DEFINE_bool(
    'kernel', False, 'Whether the input is a point cloud or kernel.')
flags.DEFINE_enum('data', '', ['branch', 'triangle', ''],
                  'Chooses simulation or user input.')
flags.DEFINE_integer('n', 3000, 'Sample size of generated data.')
flags.DEFINE_integer('p', 1000, 'Number of features in the generated data.')

# Random seeds.
flags.DEFINE_integer('k', 4, 'Seed.')
flags.DEFINE_integer('ns', 5, 'Number of seeds with which to run the model.')

# Flags for training and evaluation.
flags.DEFINE_integer('e', 10001, 'Number of epochs.')
flags.DEFINE_integer('ne', 1000, 'When to evaluate.')
flags.DEFINE_integer('nr', 1000, 'When to record the loss.')
flags.DEFINE_integer('pca', 1000, 'Applies PCA to embeddings every X epochs.')

# Flags to define the algorithm.
flags.DEFINE_boolean('keops', True, 'Uses keops or not.')
flags.DEFINE_enum('m', 'primal', ['dual', 'primal'], 'Dual or primal mode.')

# Flags for model hyperparameters.
flags.DEFINE_integer('d', 5, 'Dimension of output space.')
flags.DEFINE_string('init', 'uniform,0.,1.', 'Initialiser.')
flags.DEFINE_float('l1', 1e-4, 'Hyperparameter for penalty terms.')
flags.DEFINE_float('l2', 1e-4, 'Hyperparameter for distortion terms.')
flags.DEFINE_float('lr', 1e-5, 'Learning rate.')
flags.DEFINE_float('s', 1., 'Scale parameter.')

# Flags to assess runtime
flags.DEFINE_bool('time', False, 'Whether or not to measure runtime.')


FLAGS = flags.FLAGS


def create_dummy_config(cfg: train.ModelGetterConfig):
  """Creates dummy config, needed when timing the training loop."""
  cfg = dataclasses.replace(cfg)
  cfg.n_iter = 1
  cfg.pca = 0
  cfg.n_record = 0
  cfg.n_eval = 0
  return cfg


def time_training_loop(func):

  def inner_fn(cfg_model: train.ModelGetterConfig, *args):
    """Enables to time the training loop."""
    cfg_model_time = create_dummy_config(cfg_model)
    _ = func(cfg_model_time, *args)
    start = torch.cuda.Event(enable_timing=True)
    end = torch.cuda.Event(enable_timing=True)
    start.record()
    out = func(cfg_model, *args)
    end.record()
    # Waits for everything to finish running
    torch.cuda.synchronize()
    runtime = start.elapsed_time(end)
    return runtime, out
  return inner_fn


def main(_):
  tf.config.experimental.set_visible_devices([], 'GPU')
  logging.info('cuda is available: %s', torch.cuda.is_available())
  logging.info('size of dataset: %s x %s', FLAGS.n, FLAGS.p)

  # Gets input data.
  if FLAGS.data:
    first_view, second_view, rd_vec = data_pipeline.generate_data(
        FLAGS.n, FLAGS.p, simulation=FLAGS.data)
  else:
    first_view = data_pipeline.load(FLAGS.input_dir, FLAGS.input_fv)
    second_view = data_pipeline.load(FLAGS.input_dir, FLAGS.input_sv)
    if FLAGS.rd_vec:
      rd_vec = data_pipeline.load(FLAGS.input_dir, FLAGS.rd_vec)
    else:
      rd_vec = np.arange(first_view.shape[0])

  # Creates output directory and filename.
  gfile.makedirs(FLAGS.output_dir)
  logging.info('output directory: %s', FLAGS.output_dir)
  filename = ':'.join(['m' + str(FLAGS.m),
                       'keops' + str(FLAGS.keops),
                       'ni' + str(FLAGS.e),
                       'k' + str(FLAGS.k),
                       'ns' + str(FLAGS.ns),
                       'lr' + str(FLAGS.lr),
                       's' + str(FLAGS.s),
                       'l1' + str(FLAGS.l1),
                       'l2' + str(FLAGS.l2),
                       'i' + str(FLAGS.init)])
  if FLAGS.data:
    filename = ':'.join([filename,
                         'n' + str(FLAGS.n),
                         'p' + str(FLAGS.p),
                         str(FLAGS.data)])
  else:
    filename = ':'.join([filename,
                         FLAGS.input_fv.split('.')[0]])

  # Chooses device.
  if torch.cuda.is_available():
    device = torch.device('cuda')
  else:
    device = 'cpu'

  # Prepares metrics.
  eval_fn = SupervisedEvaluation(ground_truth_alignment=rd_vec,
                                 device=device)

  # Sets the hyperparameters of the model.
  cfg_model = train.ModelGetterConfig(
      key=FLAGS.k,
      n_seed=FLAGS.ns,
      low_dim=FLAGS.d,
      n_iter=FLAGS.e,
      keops=FLAGS.keops,
      mode=FLAGS.m,
      n_eval=FLAGS.ne,
      n_record=FLAGS.nr,
      learning_rate=FLAGS.lr,
      sigmas=FLAGS.s,
      lambda1=FLAGS.l1,
      lambda2=FLAGS.l2,
      pca=FLAGS.pca,
      init=FLAGS.init,
      )

  # Moves input to device.
  first_view = torch.FloatTensor(first_view).to(device)
  second_view = torch.FloatTensor(second_view).to(device)
  if cfg_model.mode == 'dual' and not FLAGS.kernel:
    first_view = train.get_kernel(first_view)
    second_view = train.get_kernel(second_view)

  # Runs model.
  train_fn = (time_training_loop(train.train_and_evaluate)
              if FLAGS.time else train.train_and_evaluate)
  if FLAGS.time:
    runtime, out = train_fn(
        cfg_model, first_view, second_view, eval_fn, '', device)
  else:
    out = train_fn(
        cfg_model, first_view, second_view, eval_fn, FLAGS.output_dir, device)
    runtime = '-'
  optim, model, eval_loss, eval_matching, pca_results, key = out

  # Saves results to files.
  loss = eval_loss['loss'][-1] if FLAGS.nr != 0 else -1
  mmd = eval_loss['mmd'][-1] if FLAGS.nr != 0 else -1
  foscttm = eval_matching['foscttm'][-1] if FLAGS.ne != 0 else -1
  top1 = eval_matching['top1'][-1] if FLAGS.ne != 0 else -1
  top5 = eval_matching['top5'][-1] if FLAGS.ne != 0 else -1
  results = [foscttm, top1, top5]

  logging.info('Save results in %s.', FLAGS.output_dir)
  with gfile.GFile(
      os.path.join(FLAGS.output_dir, filename + '.tsv'), 'w') as my_file:
    colnames = ['model', 'key', 'n_sample', 'n_feat', 'low_dim', 'n_iter', 'keops',
                  'loss', 'mmd', 'foscttm' ,'top1', 'top5', 'time']
    my_file.write('\t'.join(colnames) + '\n')
    checkpointer.save_data_eval(
        my_file, FLAGS, key, loss, mmd, results, runtime, cfg_model)
  if FLAGS.ne != 0:
    logging.info('Save tracking in %s.', FLAGS.output_dir)
    checkpointer.save_tracking(FLAGS.output_dir,
                               filename,
                               eval_loss,
                               eval_matching,
                               key,
                               FLAGS.e,
                               cfg_model)
  logging.info('Save model in %s.', FLAGS.output_dir)
  checkpointer.save_model(
      FLAGS.output_dir, filename, optim, model, key, FLAGS.e, loss, rd_vec)
  if FLAGS.pca != 0:
    logging.info('Save PCA representation in %s', FLAGS.output_dir)
    checkpointer.save_pca(FLAGS.output_dir, filename, pca_results, rd_vec)
  logging.info('End.')

if __name__ == '__main__':
  app.run(main)



# coding=utf-8
# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions used in MMDMA, PyTorch implementation."""
import numpy as np
import pykeops
pykeops.clean_pykeops()  # just in case old build files are still present
from pykeops.torch import LazyTensor
from sklearn import decomposition
import torch


# Functions in dual.
def pen_dual(
    embedding: torch.Tensor,
    param: torch.Tensor,
    device: torch.device
    ) -> torch.Tensor:
  """Computes the non-collapsing penalty for the dual formulation.

  In the dual formulation, when we project n points to low_dim dimensions,
  the model parameter 'param' is an n*low_dim matrix. The 'embedding'
  matrix is also an n*low_dim matrix equal to embedding=K*param,
  where K is the n*n kernel matrix of the n points in the input space.
  The 'embedding' matrix corresponds to the mapped set of points.
  The non-collapsing penalty is then equal to
  ||param.T * K * param - I ||, where I is the low_dim*low_dim
  identity matrix, which is also equal to ||param.T * embedding - I ||.

  Arguments:
    embedding: torch.Tensor, low dimensional representation of a view.
    param: torch.Tensor, model parameters.
    device: torch.device, whether cuda or cpu.

  Returns:
    penalty_value: torch.Tensor, scalar value.
  """
  low_dim = embedding.shape[1]
  identity_matrix_embedding_space = torch.eye(low_dim).to(device)
  penalty_value = (torch.matmul(param.t(), embedding)
                   - identity_matrix_embedding_space).norm(2)
  return penalty_value


def dis_dual(
    embedding: torch.Tensor,
    kernel: torch.Tensor
    ) -> torch.Tensor:
  """Computes the distortion penalty for the dual formulation.

  In the dual formulation, when we project n points to a shared space of
  low_dim dimensions, the 'embedding' matrix is an n*low_dim matrix equal to
  embedding=K*param, where K is the n*n kernel matrix of the n points in the
  input space and the n*low_dim 'param' matrix is the model parameter.  The
  'embedding' matrix corresponds to the mapped set of points.
  The distortion penalty is then equal to ||K * param * param.T * K - K ||,
  which is also equal to ||embedding * embedding.T - K||.

  Arguments:
    embedding: torch.Tensor, low dimensional representation of a view.
    kernel: torch.Tensor, view.

  Returns:
    penalty_value: torch.Tensor, scalar value.
  """
  distortion_value = (
      torch.matmul(embedding, embedding.t()) - kernel).norm(2)
  return distortion_value


# Functions for MMD.
def squared_mmd(
    first_view: torch.Tensor,
    second_view: torch.Tensor,
    sigmas: torch.Tensor,
    keops: bool,
    device: torch.device
    ) -> float:
  """Computes squared MMD with Gaussian kernel.

  Arguments:
    first_view: torch.Tensor, embedding for first view (sample x low_dim).
    second_view: torch.Tensor, embedding for second view (sample x low_dim).
    sigmas: torch.Tensor, scale parameter for Gaussian kernel.
    keops: bool, whether or not to use keops.
    device: torch.device, device can be cuda or cpu.
  Returns:
    float, MMD value.
  """
  if keops:
    cost = squared_mmd_keops(first_view, second_view, sigmas, device)
  else:
    n_sample1 = first_view.shape[0]
    n_sample2 = second_view.shape[0]
    cost = gaussian_kernel(
        first_view, first_view, sigmas, keops).sum() / n_sample1**2
    cost += gaussian_kernel(
        second_view, second_view, sigmas, keops).sum() / n_sample2**2
    cost -= 2 * gaussian_kernel(
        first_view, second_view, sigmas, keops).sum() / (
            n_sample1 * n_sample2)
  return torch.clip(cost, min=0)


def gaussian_kernel(
    first_view: torch.Tensor,
    second_view: torch.Tensor,
    sigmas: torch.Tensor,
    keops: bool
    ) -> float:
  """Computes Gaussian kernel.

  Arguments:
    first_view: torch.Tensor, embedding for first view (sample x low_dim).
    second_view: torch.Tensor, embedding for second view (sample x low_dim).
    sigmas: torch.Tensor, scale parameter of Gaussian kernel.
    keops: bool, whether or not to use keops.

  Returns:
    Gaussian kernel.
  """
  if keops:
    beta = 1.0 / (2.0 * (sigmas.unsqueeze(1)))
    m, d = first_view.shape
    n, d = second_view.shape

    first_view = first_view * torch.sqrt(beta)
    second_view = second_view * torch.sqrt(beta)
    # Turn our dense Tensors into KeOps symbolic variables with "virtual"
    # dimensions at positions 0 and 1 (for "i" and "j" indices):
    first_view_i = LazyTensor(first_view.view(m, 1, d))
    second_view_j = LazyTensor(second_view.view(1, n, d))

    # We can now perform large-scale computations, without memory overflows:
    distance_ij = ((first_view_i - second_view_j)**2).sum(dim=2)
    # Symbolic (m, n, 1) matrix of squared distances
    kernel_ij = (- distance_ij).exp()
    return kernel_ij
  else:
    beta = 1. / (2. * sigmas)
    distances = compute_sqpairwise_distances(first_view, second_view)
    kernel = torch.exp(- beta * distances)
    return kernel


def squared_mmd_keops(
    first_view: torch.Tensor,
    second_view: torch.Tensor,
    sigmas: torch.Tensor,
    device: torch.device
    ) -> float:
  """Computes the Gaussian kernel between the two given views.

  This function is inspired by https://github.com/jeanfeydy/geomloss.

  Arguments:
    first_view: torch.Tensor, first view.
    second_view: torch.Tensor, second view.
    sigmas: torch.Tensor, scale parameter for gaussian kernel.
    device: torch.device, device can be cuda or cpu.

  Returns:
    Squared MMD value.
  """
  n_sample1 = first_view.shape[0]
  n_sample2 = second_view.shape[0]
  ones1 = torch.ones(n_sample1).to(device) / n_sample1
  ones2 = torch.ones(n_sample2).to(device) / n_sample2

  # (B,N,N) tensor
  kernel_11 = gaussian_kernel(
      first_view, first_view, sigmas=sigmas, keops=True)
  # (B,M,M) tensor
  kernel_22 = gaussian_kernel(
      second_view, second_view, sigmas=sigmas, keops=True)
  # (B,N,M) tensor
  kernel_12 = gaussian_kernel(
      first_view, second_view, sigmas=sigmas, keops=True)

  # (B,N,N) @ (B,N) = (B,N)
  kernel_sum_1 = (kernel_11 @ ones1.unsqueeze(-1)).squeeze(-1)
  # (B,M,M) @ (B,M) = (B,M)
  kernel_sum_2 = (kernel_22 @ ones2.unsqueeze(-1)).squeeze(-1)
  # (B,N,M) @ (B,M) = (B,N)
  kernel_sum_12 = (kernel_12 @ ones2.unsqueeze(-1)).squeeze(-1)

  return (
      torch.dot(ones1.view(-1), kernel_sum_1.view(-1))
      + torch.dot(ones2.view(-1), kernel_sum_2.view(-1))
      - 2 * torch.dot(ones1.view(-1), kernel_sum_12.view(-1))
  )


# Functions in primal.
def pen_primal(param: torch.Tensor, device: torch.device) -> torch.Tensor:
  """Computes non-collapsing penalty for the primal formulation.

  In the primal formulation, the 'param' matrix is the n*low_dim model parameter
  and the non-collapsing penalty is equal to ||param.T * param - I||,
  where I is the low_dim*low_dim identity matrix.

  Arguments:
    param: torch.Tensor, model parameters.
    device: torch.device, whether cuda or cpu.
  Returns:
    penalty_value: torch.Tensor, scalar value.
  """
  low_dim = param.shape[1]
  identity_matrix_embedding_space = torch.eye(low_dim).to(device)
  penalty_value = (torch.matmul(param.t(), param)
                   - identity_matrix_embedding_space).norm(2)
  return penalty_value


def dis_primal(
    input_view: torch.Tensor,
    param: torch.Tensor
) -> torch.Tensor:
  """Computes distortion penalty for the primal formulation.

  In the primal formulation, the 'param' matrix is the n*low_dim model parameter
  and input_view corresponds to the input data, of shape n*p. The distortion
  penalty can be written as
  ||input_view*input_view.T - input_view*param*param.T*input_view.T|| which
  can be rewritten as sqrt(Tr((I - param*param.T)*input_view.T*input_view
  *(I - param*param.T)*input_view.T*input_view)) to
  avoid computing terms that are O(n**2) in memory or runtime.

  Arguments:
    input_view: torch.Tensor, one of the two views.
    param: torch.Tensor, model parameters.
  Returns:
    distortion_value: torch.Tensor, scalar value.
  """
  gram = torch.matmul(input_view.t(), input_view)
  tmp = torch.matmul(param, torch.matmul(param.t(), gram))
  prod = gram - tmp
  distortion_value = torch.sqrt(torch.trace(torch.matmul(prod, prod)))
  return distortion_value


# Others.
def compute_sqpairwise_distances(
    first_view: torch.Tensor,
    second_view: torch.Tensor = None
    ) -> torch.Tensor:
  """Computes squared euclidean pairwise distances.

  Args:
    first_view: numpy array, format is samples_x x features.
    second_view: numpy array, format is samples_y x features.
  Returns:
    jax numpy array, format is samples_y x samples_x.
  """
  if second_view is None:
    sample_norm_first_view = torch.sum(
        torch.square(first_view), axis=1).reshape(-1, 1)
    cross_inner_product = torch.matmul(first_view, first_view.t())
    diff = (sample_norm_first_view
            + sample_norm_first_view.reshape(1, -1) - 2 * cross_inner_product)
  else:
    sample_norm_first_view = torch.sum(
        torch.square(first_view), axis=1).reshape(-1, 1)
    sample_norm_second_view = torch.sum(
        torch.square(second_view), axis=1).reshape(1, -1)
    cross_inner_product = torch.matmul(first_view, second_view.t())
    diff = (sample_norm_first_view
            + sample_norm_second_view - 2 * cross_inner_product)
  diff = torch.clip(diff, min=0)
  return diff


# Output function.
def pca(first_view, second_view, low_dim=2):
  """Project the samples into a lower dimension with PCA."""
  pca_fn = decomposition.PCA(n_components=low_dim)
  n = first_view.shape[0]
  views = np.concatenate((first_view, second_view), axis=0)
  pca_rep = pca_fn.fit_transform(views)
  pca_rep_fv = pca_rep[:n]
  pca_rep_sv = pca_rep[n:]
  return pca_rep_fv, pca_rep_sv


